/* chrust.c : Compress a file using the Hrust algorithm
   Author: Matthew Westcott <gasman@raww.org> 2004-02-07
   The content of this file (chrust.c) is placed in the public domain.
*/

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "Z80.h"

#define min(X, Y)  ((X) < (Y) ? (X) : (Y))

static byte page[8][16384];
static int current_page = 0;

#include "hrust.h"

int
hrust(char *infile, char *outfile)
{
	struct stat infile_stat;
	FILE *f_in, *f_out;
	int fd_in, success, i, chunk;

	Z80 z;

	int compressed_length;
	int data_length;

	/* load hrust code into memory at 0x4000 (page 5) */
	memcpy(page[5], hrust_code, 2048 * sizeof(byte));

	/* load input file to 0x6000 */
	fd_in = open(infile, O_RDONLY);
	if (fd_in == -1) {
		fprintf(stderr, "Error: Could not open %s for reading\n", infile);
		return 1;
	}
	success = fstat(fd_in, &infile_stat);
	if (success != 0) {
		fprintf(stderr, "Error: Cannot stat file %s\n", infile);
		close(fd_in);
		return 1;
	}
	data_length = infile_stat.st_size;
	if (data_length > 40960) {
		fprintf(stderr, "Error: Input file is larger than 40960 bytes\n");
		close(fd_in);
		return 1;
	}
	f_in = fdopen(fd_in, "rb");
	if (f_in == NULL) {
		fprintf(stderr, "Error: Could not open %s for reading\n", infile);
		close(fd_in);
		return 1;
	}
	i = data_length;
	/* read up to 0x2000 bytes into page 5 */
	chunk = min(0x2000, i);
	if (chunk > 0) {
		fread(page[5]+0x2000, sizeof(byte), chunk, f_in);
		i -= chunk;
	}
	/* read up to 0x4000 bytes into page 2 */
	chunk = min(0x4000, i);
	if (chunk > 0) {
		fread(page[2], sizeof(byte), chunk, f_in);
		i -= chunk;
	}
	/* read up to 0x4000 bytes into page 0 */
	chunk = min(0x4000, i);
	if (chunk > 0) {
		fread(page[0], sizeof(byte), chunk, f_in);
	}
	close(fd_in);
	
	/* Set up Z80 machine state */
	ResetZ80(&z);
	z.PC.W = 16384;
	z.SP.W = 24574;
	WrZ80(24574,0); WrZ80(24575,0); /* set return address on top of stack to zero */
	/* Poke data length into the hrust routine */
	WrZ80(16389,data_length >> 8); WrZ80(16388,data_length & 0xff);
	RunZ80(&z);

	compressed_length = z.BC.W;
	printf("Compressed %s: %i => %i (%.1f%%)\n",
		infile, data_length, compressed_length,
		100.0 * compressed_length / data_length);

	f_out = fopen(outfile, "wb");
	if (f_out == NULL) {
		fprintf(stderr, "Error: Could not open %s for writing\n", outfile);
		return 1;
	}
	/* write up to 0x2000 bytes of page 5 */
	chunk = min(compressed_length, 0x2000);
	if (chunk > 0) {
		fwrite(page[5]+0x2000, sizeof(byte), chunk, f_out);
		compressed_length -= chunk;
	}
	/* write up to 0x4000 bytes of page 2 */
	chunk = min(compressed_length, 0x4000);
	if (chunk > 0) {
		fwrite(page[2], sizeof(byte), chunk, f_out);
		compressed_length -= chunk;
	}
	/* write up to 0x4000 bytes of the top page */
	chunk = min(compressed_length, 0x4000);
	if (chunk > 0) {
		fwrite(page[current_page], sizeof(byte), chunk, f_out);
	}
	fclose(f_out);
	return 0;
}

/* handlers for supporting hardware - 128K paged memory */
byte
RdZ80(register word Addr)
{
	switch (Addr & 0xc000) {
		case 0x0000: return 0;
		case 0x4000: return page[5][Addr & 0x3fff];
		case 0x8000: return page[2][Addr & 0x3fff];
		default: return page[current_page][Addr & 0x3fff];
	}
}
void
WrZ80(register word Addr,register byte Value)
{
	switch (Addr & 0xc000) {
		case 0x0000: break;
		case 0x4000: page[5][Addr & 0x3fff] = Value; break;
		case 0x8000: page[2][Addr & 0x3fff] = Value; break;
		case 0xc000: page[current_page][Addr & 0x3fff] = Value; break;
	}
}
byte
InZ80(register word Port)
{
	return 0xff;
}
void
OutZ80(register word Port,register byte Value)
{
	if ((Port & 0x8002) == 0) {
		current_page = Value & 0x07;
	}
}
word
LoopZ80(register Z80 *R)
{
	/* exit emulation when PC returns to below 0x4000 */
	return (R->PC.W & 0xc000 ? INT_NONE : INT_QUIT);
}
void
PatchZ80(register Z80 *R)
{
	/* no patching necessary */
}

int
main(int argc, char **argv)
{
	if (argc != 3) {
		fprintf(stderr, "Usage: %s input_file output_file\n", argv[0]);
		return 1;
	}

	return hrust(argv[1], argv[2]);
}
